/*
 * xhtmlbuffer.c
 * xhtmlite
 *
 * by Doug Wolfe
 * doug@wolfescience.com 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc.,  59 Temple Place - Suite 330, Cambridge, MA 02139, USA.
 *
 */

/* TODO (#2#): image resize */
/* TODO (#1#): add tags for font family */
/* TODO (#1#): track tags and show if they are active using the 
               buttons */
/* TODO (#1#): Fix import html with image (maybe losing file information 
               on the original loading of the file); */

#include <gtk/gtk.h>

#include "xhtmlinterface.h"
#include "xhtmlbuffer.h"
#include "xhtmltag.h"
#include "support.h"
#include "base64.h"

#include <string.h>
#include <gdk-pixbuf/gdk-pixdata.h>

//for xml
#include <libxml/xmlreader.h>
#include <libxml/encoding.h>
#include <libxml/xmlwriter.h>

#define MY_ENCODING "UTF-8"


static GtkTextTagTable * tagtable;

static GList * undo_stack = NULL;
static GList * redo_stack = NULL;

static gint buffer_changed_id;

static GdkPixbuf * byoj_image_repr_read_b64 (const gchar * uri_data)
{
	GdkPixbuf * pixbuf = NULL;
	GdkPixbufLoader * loader = NULL;
 
	gint j;
	gint k;
	gint l;
	gint b;
	gint len;
	gint eos = 0;
	gint failed = 0;
 
	guint32 bits;
 
	static const gchar B64[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
	const gchar* btr = uri_data;
 
	gchar ud[4];
 
	guchar bd[57];
 
	loader = gdk_pixbuf_loader_new ();
 
	if (loader == NULL) return NULL;
 
	while (eos == 0) {
		l = 0;
		for (j = 0; j < 19; j++) {
			len = 0;
			for (k = 0; k < 4; k++) {
				while (isspace ((int) (*btr))) {
					if ((*btr) == '\0') break;
					btr++;
				}
				if (eos) {
					ud[k] = 0;
					continue;
				}
				if (((*btr) == '\0') || ((*btr) == '=')) {
					eos = 1;
					ud[k] = 0;
					continue;
				}
				ud[k] = 64;
				for (b = 0; b < 64; b++) { /* There must a faster way to do this... ?? */
					if (B64[b] == (*btr)) {
						ud[k] = (gchar) b;
						break;
					}
				}
				if (ud[k] == 64) { /* data corruption ?? */
					eos = 1;
					ud[k] = 0;
					continue;
				}
				btr++;
				len++;
			}
			bits = (guint32) ud[0];
			bits = (bits << 6) | (guint32) ud[1];
			bits = (bits << 6) | (guint32) ud[2];
			bits = (bits << 6) | (guint32) ud[3];
			bd[l++] = (guchar) ((bits & 0xff0000) >> 16);
			if (len > 2) {
				bd[l++] = (guchar) ((bits & 0xff00) >>  8);
			}
			if (len > 3) {
				bd[l++] = (guchar)  (bits & 0xff);
			}
		}
 
		if (!gdk_pixbuf_loader_write (loader, (const guchar *) bd, (size_t) l, NULL)) {
			failed = 1;
			break;
		}
	}
 
	gdk_pixbuf_loader_close (loader, NULL);
	if (!failed) pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);
	return pixbuf;
}

static void xhtml_add_style_font_color(GtkTextBuffer * textbuffer,
				       const gchar * color_string,
				       const GtkTextIter *start,
				       const GtkTextIter *end)
{
	GtkTextTag * tag;
	GdkColor color;
	color_string = g_utf8_strdown(color_string, -1);
	if(gdk_color_parse(color_string, &color))
	{
			//make string and convert color to hexadecimal
		gchar * color_string =g_strdup_printf("span color='%s'", color_string);
		printf("%s\n", color_string);

		tag = gtk_text_tag_table_lookup(tagtable, color_string);
						//look to see if tag exists and create it if not
		if(tag == NULL)
		{
			tag = gtk_text_buffer_create_tag (textbuffer, color_string, NULL);
			setup_tag (tag);
			g_object_set (tag, "foreground_gdk", &color, NULL);
		}
		if(tag != NULL)
		{
			gtk_text_buffer_apply_tag(textbuffer, tag, start, end);
		}
	}
	else
	{
		printf("color did not parse correctly:%s\n", color_string);
	}
}
static void xhtml_add_style_font_background_color(GtkTextBuffer * textbuffer,
					const gchar * color_string,
				       const GtkTextIter *start,
				       const GtkTextIter *end)
{
	GtkTextTag * tag;
	GdkColor color;
	color_string = g_utf8_strdown(color_string, -1);
	if(gdk_color_parse(color_string, &color))
	{
		//make string and convert color to hexadecimal
		gchar * color_string =g_strdup_printf("span background-color='%s'", attribute);
		printf("color_string:%s\n", color_string);

		tag = gtk_text_tag_table_lookup(tagtable, color_string);
						//look to see if tag exists and create it if not
		if(tag == NULL)
		{
			tag = gtk_text_buffer_create_tag (textbuffer, color_string, NULL);
			setup_tag (tag);
			g_object_set (tag, "background_gdk", &color, NULL);
		}
		if(tag != NULL)
		{
			gtk_text_buffer_apply_tag(textbuffer, tag, start, end);
		}
	}
	else
	{
		printf("color did not parse correctly:%s\n", attribute);
	}
}

static void xhtml_add_style_font_face(GtkTextBuffer * textbuffer,
				      const gchar * face,
				       const GtkTextIter *start,
				       const GtkTextIter *end)
{
	GtkTextTag * tag;
	//create the font-size tag name
	//do not lowercase this attribute as font face is case sensitive. For instance: "Sans"
	gchar * fontfamily = g_strdup_printf("span face='%s'", face);
	printf("fontfamily:%s\n", fontfamily);
					//look to see if tag exists and create if not
	tag = gtk_text_tag_table_lookup(tagtable, fontfamily);
	if(tag == NULL)
	{
		printf("fontfamily tag not found: %s\n", fontfamily);
		tag = gtk_text_buffer_create_tag (textbuffer, fontfamily, NULL);
		setup_tag (tag);
		g_object_set (tag,
			      "family", attribute,
			      NULL);
	}
	if(tag != NULL)
	{
		gtk_text_buffer_apply_tag(textbuffer, tag, start, end);
	}
}

static void xhtml_add_style_font_size(GtkTextBuffer * textbuffer,
				      const gchar * font_size,
				      const GtkTextIter *start,
				      const GtkTextIter *end)
{
	GtkTextTag * tag;
	font_size = g_utf8_strdown(font_size, -1);
	//create the font-size tag name
	gchar * fontsize_string =g_strdup_printf("span font-size='%s'", font_size);
	printf("fontsize_string:%s\n", fontsize_string);
					//look to see if tag exists and release error if it is not found
	tag = gtk_text_tag_table_lookup(tagtable, fontsize_string);
	if(tag == NULL)
	{
		printf("font-size tag not found: %s\n", fontsize_string);
	}
	if(tag != NULL)
	{
		gtk_text_buffer_apply_tag(textbuffer, tag, start, end);
	}
}

static void process_font_tag(GtkTextBuffer * textbuffer, xmlTextReaderPtr reader,
			     const GtkTextIter *start,
			     const GtkTextIter *end)
{
	const gchar * attribute;
	attribute = (const gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"color");
	if(attribute != NULL)
	{
		xhtml_add_style_font_color(textbuffer, attribute, start, end);
	}
	attribute = (const gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"background-color");
	if(attribute != NULL)
	{
		xhtml_add_style_font_background_color(textbuffer, attribute, start, end);
	}
	attribute = (const gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"font-size");
	if(attribute != NULL)
	{
		xhtml_add_style_font_size(textbuffer, attribute, start, end);
	}
	attribute = (const gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"face");
	if(attribute != NULL)
	{
		xhtml_add_style_font_face(textbuffer, attribute, start, end);
	}
}

static void parse_and_process_style_attribute(GtkTextBuffer * textbuffer, xmlTextReaderPtr reader,
			     const GtkTextIter *start,
			     const GtkTextIter *end)
{
	GtkTextTag * tag;
	const gchar * attribute;
	attribute = (const gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"style");
	if(attribute != NULL)
	{
		gchar * font_size;
		gchar * font_face;
		gchar * color;
		gchar * background_color;

		gchar** split_style = g_strsplit_set (attribute, ";",-1);
		int n;
		for (n = 0; split_style[n] ; n++)
		{
			printf("split_style(%d):%s\n", n, split_style[n]);
			gchar** split_substyle = g_strsplit_set (split_style[n], ":",-1);
			printf("split_style(%d):%s\n", 0, split_substyle[0]);
			printf("split_style(%d):%s\n", 1, split_substyle[1]);

			if(strcmp(split_substyle[0], "color") == 0)
			{
				xhtml_add_style_font_color(textbuffer, split_substyle[1], start, end);
			}
			if(strcmp(split_substyle[0], "background-color") == 0)
			{
				xhtml_add_style_font_background_color(textbuffer, split_substyle[1], start, end);
			}
			if(strcmp(split_substyle[0], "font-size") == 0)
			{
				xhtml_add_style_font_size(textbuffer, split_substyle[1], start, end);
			}
			if(strcmp(split_substyle[0], "face") == 0)
			{
				xhtml_add_style_font_face(textbuffer, split_substyle[1], start, end);
			}
		}
	}
}

static gboolean TagEndsHere (GtkTextTag * tag, GtkTextIter * iter, GtkTextIter * next_iter)
{
	return (gtk_text_iter_has_tag (iter, tag) && !gtk_text_iter_has_tag (next_iter, tag) || gtk_text_iter_is_end(next_iter));
}

static gboolean toggle_selected_tag (GtkTextBuffer *textbuffer, gchar * tag_name)
{
	GtkTextIter beginselect, endselect;
	gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);
	GtkTextTag * tag = gtk_text_tag_table_lookup(tagtable, tag_name);

	while(!gtk_text_iter_equal(&beginselect, &endselect))
	{
		if(!gtk_text_iter_has_tag (&beginselect, tag))
		{
			gtk_text_buffer_apply_tag (textbuffer, tag, &beginselect, &endselect);
			return (TRUE);
		}
		gtk_text_iter_forward_char(&beginselect);
	}
	gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);
	gtk_text_buffer_remove_tag (textbuffer, tag, &beginselect, &endselect);
	return (FALSE);
}

static gboolean set_selected_tag (GtkTextBuffer *textbuffer, gchar * tag_name, gboolean enable)
{
	GtkTextIter beginselect, endselect;
	gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);
	GtkTextTag * tag = gtk_text_tag_table_lookup(tagtable, tag_name);

	if(enable)
	{
		gtk_text_buffer_apply_tag (textbuffer, tag, &beginselect, &endselect);
		return (TRUE);
	}

	gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);
	gtk_text_buffer_remove_tag (textbuffer, tag, &beginselect, &endselect);
	return (FALSE);
}

static gchar * xhtml_append_style (gchar * style_string, gchar * style_name, gchar * style_value)
{
	if(style_string==NULL)
	{
		return g_strdup_printf("%s:%s", style_name, style_value);
	}
	else
	{
		return g_strdup_printf("%s;%s:%s", style_string, style_name, style_value);
	}
}
	
xmlChar *ConvertXmlInput(const char *in, const char *encoding);

void WriteStartElement(xmlTextWriterPtr writer, GtkTextTag * tag)
{
	//write start xml element
	int rc;
	if(g_strrstr(tag->name, "span\0")!= NULL)
	{
		//set font with attributes
		rc = xmlTextWriterStartElement(writer, BAD_CAST "span");
		if (rc < 0) {
			printf	("xmlTextWriterStartElement: Error writing 'span' element\n");
			return;
		}
		
		gchar * style = NULL;
		gchar * value;
		if(g_strrstr(tag->name, "face\0")!= NULL)
		{
			g_object_get (tag, "family", &value, NULL);
			style = xhtml_append_style (style, "face", value);
		}
		
		if(g_strrstr(tag->name, "font-size\0")!= NULL)
		{
			value = g_object_get_data(G_OBJECT(tag), "font-size_absolute");
			style = xhtml_append_style (style, "font-size", value);
		}
	
		if(g_strrstr(tag->name, "color\0")!= NULL)
		{
		//make string and convert color to hexadecimal
			GdkColor *color;
			if (g_strrstr(tag->name, "background-color")!= NULL)
			{
				g_object_get (tag, "background_gdk", &color, NULL);
				value = g_strdup_printf("#%x%x%x", color->red/256, color->green/256, color->blue/256);
				style = xhtml_append_style (style, "background-color", value);
			}
			else if (g_strrstr(tag->name, "color")!= NULL)
			{
				g_object_get (tag, "foreground_gdk", &color, NULL);
				value = g_strdup_printf("#%x%x%x", color->red/256, color->green/256, color->blue/256);
				style = xhtml_append_style (style, "color", value);
			}
		}

		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "style",
				BAD_CAST style);
		if (rc < 0) {
			printf("xmlTextWriterWriteAttribute: Error writing 'face' attribute\n");
			return;
		}
	}
	else if(g_strrstr(tag->name, "font\0")!= NULL)
	{
		//set font with attributes
		rc = xmlTextWriterStartElement(writer, BAD_CAST "font");
		if (rc < 0) {
			printf	("xmlTextWriterStartElement: Error writing 'font' element\n");
			return;
		}

		if(g_strrstr(tag->name, "face\0")!= NULL)
		{
			gchar * fontfamily;
			g_object_get (tag, "family", &fontfamily, NULL);
			rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "face",
					BAD_CAST fontfamily);
			if (rc < 0) {
				printf("xmlTextWriterWriteAttribute: Error writing 'face' attribute\n");
				return;
			}
		}
		
		if(g_strrstr(tag->name, "font-size\0")!= NULL)
		{
			gchar * fontsize = g_object_get_data(G_OBJECT(tag), "font-size_absolute");
			rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "font-size",
					BAD_CAST fontsize);
			if (rc < 0) {
				printf("xmlTextWriterWriteAttribute: Error writing absolute 'font-size'\n");
				return;
			}
		}
	
		if(g_strrstr(tag->name, "color\0")!= NULL)
		{
		//make string and convert color to hexadecimal
			GdkColor *color;
			gchar *color_string;
			if (g_strrstr(tag->name, "background-color")!= NULL)
			{
				g_object_get (tag, "background_gdk", &color, NULL);
				color_string = g_strdup_printf("#%x%x%x", color->red/256, color->green/256, color->blue/256);
				printf("background-color=%s\n", color_string);
				/* Add an attribute with name "version" and value "1.0" to ORDER. */
				rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "background-color",
						BAD_CAST color_string);
				if (rc < 0) {
					printf("xmlTextWriterWriteAttribute: Error writing 'background-color' tag\n");
					return;
				}
			}
			else if (g_strrstr(tag->name, "color")!= NULL)
			{
				g_object_get (tag, "foreground_gdk", &color, NULL);
				color_string = g_strdup_printf("#%x%x%x", color->red/256, color->green/256, color->blue/256);
				printf("color=%s\n", color_string);
				/* Add an attribute with name "version" and value "1.0" to ORDER. */
				rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "color",
						BAD_CAST color_string);
				if (rc < 0) {
					printf("xmlTextWriterWriteAttribute: Error writing 'color' tag\n");
					return;
				}
			}
		}
	}
	else
	{
		rc = xmlTextWriterStartElement(writer, BAD_CAST tag->name);
		if (rc < 0) {
			printf	("xmlTextWriterStartElement: Error at writing tag '%s'\n", tag->name);
			return;
		}
	}
}

void
WriteEndElement(xmlTextWriterPtr writer, GtkTextTag * tag)
{
	/* Close the element */
	int rc;
	rc = xmlTextWriterEndElement(writer);
	if (rc < 0) {
		printf("testXmlwriterFilename: Error at xmlTextWriterEndElement\n");
		return;
	}
}

/**
* ConvertInput:
 * @in: string in a given encoding
 * @encoding: the encoding used
 *
 * Converts @in into UTF-8 for processing with libxml2 APIs
 *
 * Returns the converted UTF-8 string, or NULL in case of error.
 */
xmlChar * ConvertXmlInput(const char *in, const char *encoding)
{
	xmlChar *out;
	int ret;
	int size;
	int out_size;
	int temp;
	xmlCharEncodingHandlerPtr handler;

	if (in == 0)
		return 0;

	handler = xmlFindCharEncodingHandler(encoding);

	if (!handler) {
		printf("ConvertInput: no encoding handler found for '%s'\n",
		       encoding ? encoding : "");
		return 0;
	}

	size = (int) strlen(in) + 1;
	out_size = size * 2 - 1;
	out = (unsigned char *) xmlMalloc((size_t) out_size);

	if (out != 0) {
		temp = size - 1;
		ret = handler->input(out, &out_size, (const xmlChar *) in, &temp);
		if ((ret < 0) || (temp - size + 1)) {
			if (ret < 0) {
				printf("ConvertInput: conversion wasn't successful.\n");
			} else {
				printf
						("ConvertInput: conversion wasn't successful. converted: %i octets.\n",
						temp);
			}

			xmlFree(out);
			out = 0;
		}
		else {
			out = (unsigned char *) xmlRealloc(out, out_size + 1);
			out[out_size] = 0;  /*null terminating out */
		}
	} else {
		printf("ConvertInput: no mem\n");
	}

	return out;
}

gchar * WriteXHTML_as_Markup (GtkTextBuffer *textbuffer,
				      const char *file)
{
	gchar * XHTML = WriteXHTML(textbuffer, file);
	strrepl(XHTML, "<html>", "", XHTML);
	strrepl(XHTML, "</html>\n", "", XHTML);
	strrepl(XHTML, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n", "", XHTML);
	return XHTML;
}

    static guint    gen_type = GDK_PIXDATA_DUMP_PIXDATA_STREAM;
    static guint    gen_ctype = GDK_PIXDATA_DUMP_GTYPES | GDK_PIXDATA_DUMP_STATIC | GDK_PIXDATA_DUMP_CONST;
  
    static gboolean use_rle = TRUE;
    static gboolean with_decoder = FALSE;
    static gchar   *image_name = "my_pixbuf";
#define GDK_PIXBUF_MAGIC_NUMBER (0x47646b50)    /* 'GdkP' */
gchar * WriteXHTML (GtkTextBuffer *textbuffer, const char *file)
{
	int rc;
	xmlTextWriterPtr writer;
	xmlBufferPtr buf;
	xmlChar *tmp;
	FILE *fp;
	//TODO: for image writing: xmlTextWriterWriteBase64	
	printf("Started write of xml.\n");
    /* Create a new XML buffer, to which the XML document will be
    * written */
	buf = xmlBufferCreate();
	if (buf == NULL) {
		printf("testXmlwriterMemory: Error creating the xml buffer\n");
		return;
	}

    /* Create a new XmlWriter for memory, with no compression.
    * Remark: there is no compression for this kind of xmlTextWriter */
	writer = xmlNewTextWriterMemory(buf, 0);
	if (writer == NULL) {
		printf("testXmlwriterMemory: Error creating the xml writer\n");
		return;
	}

	/* Start the document with the xml default for the version*/
	rc = xmlTextWriterStartDocument(writer, NULL, MY_ENCODING, NULL);
	if (rc < 0) {
		printf("testXmlwriterMemory: Error at xmlTextWriterStartDocument\n");
		return;
	}

	/* Start an element named "HTML". Since thist is the first
	* element, this will be the root element of the document. */
	printf("write start document element\n");
	rc = xmlTextWriterStartElement(writer, BAD_CAST "html");
	if (rc < 0) {
		printf("testXmlwriterMemory: Error at xmlStartElement\n");
		return;
	}

	GList * tag_stack = NULL;
	GList * replay_stack = NULL;

	GtkTextIter start, end;
	gtk_text_buffer_get_start_iter (textbuffer, &start);
	gtk_text_buffer_get_end_iter (textbuffer, &end);

	GtkTextIter iter, next_iter;
	gtk_text_buffer_get_start_iter (textbuffer, &iter);
	gtk_text_buffer_get_start_iter (textbuffer, &next_iter);
	gtk_text_iter_forward_char(&next_iter);

	GSList * tags =  gtk_text_iter_get_tags(&start);
	int n;
	guint len =  g_slist_length(tags);
	printf ("start tags len:%d\n", len);
	for(n = 0; n < len; n++)
	{
		GtkTextTag * start_tag = g_slist_nth_data (tags, n);
		//if (TVTagTable.TagIsIgnored (start_tag))
 		//	continue;
  
		if (!gtk_text_iter_toggles_tag(&start, start_tag))
		{
 			//tag_stack.Push (start_tag);
			tag_stack=g_list_prepend (tag_stack, start_tag); //push iter onto the beginning of the list
			WriteStartElement(writer, start_tag);
			//xml.WriteStartElement (TagXmlStart(start_tag)); 
			
 			//TagXmlAttribute(start_tag, xml);
		}
	}

	while (!gtk_text_iter_equal(&iter, &end) && !gtk_text_iter_is_end(&iter))
	{
		tags =  gtk_text_iter_get_tags(&iter);
		len =  g_slist_length(tags);
		for(n = 0; n < len; n++)
		{
			GtkTextTag * tag = g_slist_nth_data (tags, n);

				//if (TVTagTable.TagIsIgnored (start_tag))
				//	continue;
			if (gtk_text_iter_begins_tag(&iter, tag))
			{
				//tag_stack.Push (start_tag);
				tag_stack = g_list_prepend (tag_stack, tag); //push iter on to the beginning of the list
				printf("tag added\n");
				WriteStartElement(writer, tag);
				//TagXmlAttribute(start_tag, xml);
			}
		}
 		// Hidden character representing an anchor
		gunichar iter_char = gtk_text_iter_get_char(&iter);
		if (iter_char ==  0xFFFC) //for a not valid unicode characters, such as an image
		{
			printf("Got child anchor!!!\n");
 		
			GdkPixbuf *mypixbuf = gtk_text_iter_get_pixbuf (&iter);
		
			//get file_string
			if(mypixbuf != NULL)
			{
				printf("write start image element\n");
				rc = xmlTextWriterStartElement(writer, BAD_CAST "img");
				if (rc < 0) {
					printf("testXmlwriterMemory: Error at xmlStartElement\n");
					return;
				}
		
				xmlChar * file_string= (xmlChar*)g_object_get_data(G_OBJECT(mypixbuf), "file_string");
		;
				//printf("file_string:%s\n", file_string);
				//xmlTextWriterWriteFormatAttribute does not seem to work here
				rc = xmlTextWriterStartAttribute(writer, BAD_CAST "src");
				if (rc < 0) {
				//printf("testXmlwriterFilename: Error at xmlTextWriterWriteAttribute\n");
					return;
				}
				rc = xmlTextWriterWriteString(writer, BAD_CAST "data:image;base64,");
				if (rc < 0) {
				//printf("testXmlwriterFilename: Error at xmlTextWriterWriteAttribute\n");
					return;
				}
				rc = xmlTextWriterWriteString(writer, BAD_CAST file_string);
				if (rc < 0) {
				//printf("testXmlwriterFilename: Error at xmlTextWriterWriteAttribute\n");
					return;
				}
				rc = xmlTextWriterEndAttribute	(writer);
				if (rc < 0) {
				//printf("testXmlwriterFilename: Error at xmlTextWriterWriteAttribute\n");
					return;
				}
				rc = xmlTextWriterEndElement(writer);
				if (rc < 0) {
				//printf("testXmlwriterFilename: Error at xmlTextWriterEndElement\n");
					return;
				}
			}

			// if (iter.ChildAnchor != null)
			// {
			// 	//Console.WriteLine(iter.ChildAnchor.ToString());
			// 	string serialize = (string) iter.ChildAnchor.Data["serialize"];
			// 	if (serialize != null)
			// 		xml.WriteRaw (serialize);
			// }
			
		}
		else
		{
			//tmp = ConvertInput(BAD_CAST &iter_char, MY_ENCODING);
			static gchar charbuf[10];
			int charlen;
			charlen = g_unichar_to_utf8 (iter_char, charbuf);
			charbuf[charlen] = '\0';
			xmlTextWriterWriteString(writer, BAD_CAST charbuf);
		}

		for(n = 0; n < len; n++)
		{
			GtkTextTag *tag = g_slist_nth_data (tags, n);

			//if (TVTagTable.TagIsIgnored (tag))
			//	continue;
			if (TagEndsHere (tag, &iter, &next_iter))
			{
				// Unwind until the ended tag.
				// Put any intermediate tags
				// into a replay queue...
				printf ("tag_stack_len:%d\n", g_list_length(tag_stack));
				while (g_list_length(tag_stack) > 0)
				{
					GtkTextTag * existing_tag = g_list_nth_data(tag_stack, 0);
					tag_stack=g_list_remove(tag_stack, existing_tag);
					if(!TagEndsHere (existing_tag, &iter, &next_iter))
					{
						replay_stack=g_list_prepend (replay_stack, existing_tag); //push iter on to the list the beginning of the list
					}
					/* Close the element */
					WriteEndElement(writer, tag);
					printf ("tag_stack_len:%d\n", g_list_length(tag_stack));
				}
					// Replay the replay queue.
					// Restart any tags that
					// overlapped with the ended
					// tag...
				while (g_list_length(replay_stack) > 0)
				{
					GtkTextTag * replay_tag = g_list_nth_data(replay_stack, 0);
					replay_stack= g_list_remove(replay_stack, replay_tag);
					tag_stack=g_list_prepend (tag_stack, replay_tag); //push iter on to the list the beginning of the list
					WriteStartElement(writer, replay_tag);
					//TagXmlAttribute(replay_tag, xml);
				}
			}
		}


		if (gtk_text_iter_ends_line(&iter))
		{
			/* Start an element named "br" for linebreak. */
			rc = xmlTextWriterStartElement(writer, BAD_CAST "br");
			if (rc < 0) {
				printf	("testXmlwriterFilename: Error at xmlTextWriterStartElement\n");
				return;
			}
			/* Close the element named "br" for linebreak. */
			rc = xmlTextWriterEndElement(writer);
			if (rc < 0) {
				printf("testXmlwriterFilename: Error at xmlTextWriterEndElement\n");
				return;
			}
		}
		gtk_text_iter_forward_char(&iter);
		gtk_text_iter_forward_char(&next_iter);
	}

	/* Here we could close the elements and since we do not want to
	* write any other elements, we simply call xmlTextWriterEndDocument,
	* which will do all the work. */
	rc = xmlTextWriterEndDocument(writer);
	if (rc < 0) {
		printf("testXmlwriterMemory: Error at xmlTextWriterEndDocument\n");
		return;
	}

	xmlFreeTextWriter(writer);

	gchar * charbuffer= g_strdup((const gchar *)buf->content);
	//printf("xml:%s", charbuffer);

	//gtk_text_buffer_set_text (buffer_output, charbuffer, -1);

	xmlBufferFree(buf);
	return charbuffer;
}

GtkTextBuffer * xhtmlite_create_buffer (GtkTextView *textview, gboolean writeable)
{
	//create and set buffer
	//add if for tag table already created
	tagtable =  gtk_text_tag_table_new();
	GtkTextBuffer * textbuffer = gtk_text_buffer_new (tagtable);
	fill_tagtable(textbuffer);
	gtk_text_view_set_buffer(textview, textbuffer);

	add_undo(textbuffer);

	if(writeable)
	{
		buffer_changed_id = g_signal_connect ((gpointer) textbuffer, "changed",
				G_CALLBACK (on_buffer_changed),
				NULL);
	}
	return textbuffer;
}

void set_XHTML (GtkTextBuffer *textbuffer, gchar *xml)
{
	GtkTextIter beginselect, endselect;
	gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);

	//create and set buffer
	gtk_text_buffer_delete  (textbuffer, &beginselect, &endselect);
	undo_stack = NULL;
	redo_stack = NULL;

	import_XHTML (textbuffer, xml, FALSE);
}

void on_buffer_changed (GtkTextBuffer * textbuffer, gpointer user_data)
{
	add_undo(textbuffer);
}

//TODO: Add Remember selection beginning and end to U
void add_undo (GtkTextBuffer * textbuffer)
{
	undo_stack=g_list_prepend (undo_stack, WriteXHTML(textbuffer, "filename"));
}

void add_redo (GtkTextBuffer * textbuffer)
{
	redo_stack=g_list_prepend (redo_stack, WriteXHTML(textbuffer, "filename"));
}

void Undo (GtkTextBuffer *textbuffer)
{
	printf("Start UNDO\n");
	g_signal_handler_disconnect ((gpointer) textbuffer, buffer_changed_id);
	
	gchar * undochar = g_list_nth_data (undo_stack, 1);
	if (undochar != NULL)
	{
		printf("UNDO:%s\n", undochar);
		add_redo (textbuffer);
		import_XHTML  (textbuffer, undochar, FALSE);
		gchar * oldundo = g_list_nth_data (undo_stack, 0);
		undo_stack = g_list_remove(undo_stack, oldundo);
	}
	buffer_changed_id = g_signal_connect ((gpointer) textbuffer, "changed",
					       G_CALLBACK (on_buffer_changed),
					       NULL);
}

void Redo (GtkTextBuffer * textbuffer)
{
	printf("Start REDO\n");
	g_signal_handler_disconnect ((gpointer) textbuffer, buffer_changed_id);
	
	gchar * redochar = g_list_nth_data (redo_stack, 0);
	if (redochar != NULL)
	{
		printf("REDO:%s\n", redochar);
		
		import_XHTML  (textbuffer, redochar, FALSE);
		redo_stack = g_list_remove(redo_stack, redochar);
		add_undo(textbuffer);
	}

	buffer_changed_id = g_signal_connect ((gpointer) textbuffer, "changed",
					       G_CALLBACK (on_buffer_changed),
					       NULL);
}

void toggle_Bold (GtkTextBuffer * textbuffer, gboolean enable)
{
	set_selected_tag (textbuffer, "b", enable);
}

void toggle_Italic (GtkTextBuffer * textbuffer, gboolean enable)
{
	set_selected_tag (textbuffer, "i", enable);
}

void toggle_Underline (GtkTextBuffer * textbuffer, gboolean enable)
{
	set_selected_tag (textbuffer, "u", enable);
}

void toggle_Strikethrough (GtkTextBuffer * textbuffer, gboolean enable)
{
	set_selected_tag (textbuffer, "del", enable); //Strikethrough
}

void toggle_Superscript (GtkTextBuffer * textbuffer, gboolean enable)
{
	GtkTextIter beginselect, endselect;
	gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "sub", &beginselect, &endselect);
	set_selected_tag(textbuffer, "sup", enable);
}

void toggle_Subscript (GtkTextBuffer * textbuffer, gboolean enable)
{
	GtkTextIter beginselect, endselect;
	gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "sup", &beginselect, &endselect);
	set_selected_tag(textbuffer, "sub", enable);
}

void adjust_fontsize_absolute(GtkTextBuffer *textbuffer, gchar * fontsize)
{
	GtkTextIter beginselect, endselect;
	gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='xx-small'", &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='x-small'", &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='small'", &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='medium'", &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='large'", &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='x-large'", &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='xx-large'", &beginselect, &endselect);
	
	toggle_selected_tag(textbuffer, g_strdup_printf("span font-size='%s'", fontsize));
}

void xhtmlite_change_fontfamily (GtkTextBuffer * textbuffer, GtkWidget * widget, gchar * fontfamily)
{
	GtkTextIter beginselect, endselect;
	gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);
	gchar * fontfamily_tagname;
	GtkTextTag *tag;
	//TODO: Maybe, Use combobox instead to get list of fonts
	//remove all old fontfamily tags
	PangoContext * textview_context = gtk_widget_get_pango_context(widget);
	PangoFontFamily **families;
	int families_len, n;
	pango_context_list_families (textview_context, &families, &families_len);
	for (n = 0;  n < families_len;  n++)
	{
		const gchar * fontname =  pango_font_family_get_name (families[n]);
		fontfamily_tagname = g_strdup_printf("span face='%s'", fontname);
		tag = gtk_text_tag_table_lookup(tagtable, fontfamily_tagname);
		if (tag != NULL)
		{	
			gtk_text_buffer_remove_tag(textbuffer, tag, &beginselect, &endselect);
		}
	}

	//new tag
	fontfamily_tagname = g_strdup_printf("span face='%s'", fontfamily);
	tag = gtk_text_tag_table_lookup(tagtable, fontfamily_tagname);
	//look to see if tag exists and if doesn't exist, create it 
	if(tag == NULL)
	{
		tag = gtk_text_buffer_create_tag (textbuffer, fontfamily_tagname, NULL);
		setup_tag (tag);
		g_object_set (tag,
			      "family", fontfamily,
			      NULL);
	}
	
	gtk_text_buffer_apply_tag (textbuffer, tag, &beginselect, &endselect);
}

void open_font_color_dialog (GtkTextBuffer * textbuffer)
{
	GtkWidget * dialog;
	dialog = create_colorselectiondialog1();

	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK)
	{
		GtkWidget * color_selection1 = lookup_widget(GTK_WIDGET(dialog), "color_selection1");
		GdkColor color;
		gtk_color_selection_get_current_color(GTK_COLOR_SELECTION(color_selection1), &color);
    
	//make string and convert color to hexadecimal
		gchar  * color_string = g_strdup_printf("span color='#%x%x%x'", color.red/256, color.green/256, color.blue/256);
		printf("%s\n", color_string);

		GtkTextTag *tag = gtk_text_tag_table_lookup(tagtable, (gchar *)color_string);
	//look to see if tag exists and create it if not
		if(tag == NULL)
		{
			tag = gtk_text_buffer_create_tag (textbuffer, (gchar *)color_string, NULL);
			setup_tag (tag);
			g_object_set (tag,
				      "foreground_gdk", &color,
				      NULL);
		}
	
		GtkTextIter beginselect, endselect;
		gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);
		gtk_text_buffer_apply_tag (textbuffer, tag, &beginselect, &endselect);
	}
	gtk_widget_hide(dialog);
	
	
	gtk_widget_destroy (dialog);
}

void open_highlight_color_dialog (GtkTextBuffer * textbuffer)
{
	GtkWidget * dialog;
	dialog = create_colorselectiondialog1();

	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK)
	{
		GtkWidget * color_selection1 = lookup_widget(GTK_WIDGET(dialog), "color_selection1");
		GdkColor color;
		gtk_color_selection_get_current_color(GTK_COLOR_SELECTION(color_selection1), &color);
    
	//make string and convert color to hexadecimal
		gchar  * color_string = g_strdup_printf("span background-color='#%x%x%x'", color.red/256, color.green/256, color.blue/256);
		printf("%s\n", color_string);

		GtkTextTag *tag = gtk_text_tag_table_lookup(tagtable, (gchar *)color_string);
	//look to see if tag exists and create it if not
		if(tag == NULL)
		{
			tag = gtk_text_buffer_create_tag (textbuffer, (gchar *)color_string, NULL);
			setup_tag (tag);
			g_object_set (tag,
				      "background_gdk", &color,
				      NULL);
		}
	
		GtkTextIter beginselect, endselect;
		gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);
		gtk_text_buffer_apply_tag (textbuffer, tag, &beginselect, &endselect);
	}
	gtk_widget_hide(dialog);
	
	
	gtk_widget_destroy (dialog);
}

GList * import_tag_stack = NULL;

int offset;
void import_XHTML (GtkTextBuffer * textbuffer, gchar * xml, gboolean markup)
{
	GtkTextIter begin, end;
	gtk_text_buffer_get_bounds (textbuffer, &begin, &end);
	gtk_text_buffer_delete(textbuffer, &begin, &end);

	//Fix White Spaces
	strrepl(xml, "\r", "", xml);
	strrepl(xml, "\n", "", xml);
	
	strrepl(xml, "<br />", "\n", xml);//the correct xhtml version
	strrepl(xml, "<br>", "\n", xml); //html version
	strrepl(xml, "<br></br>", "\n", xml); //incorrect xhtml
	strrepl(xml, "<br/>", "\n", xml);//incorrect xhtml
	
	strrepl(xml, "<BR />", "\n", xml);//the correct xhtml version
	strrepl(xml, "<BR>", "\n", xml); //html version
	strrepl(xml, "<BR></BR>", "\n", xml); //incorrect xhtml
	strrepl(xml, "<BR/>", "\n", xml);//incorrect xhtml

	gchar  * formatted_xml;
	if(markup)
	{
		formatted_xml = g_strdup_printf("<html>%s</html>", xml);
	}
	else
	{
		formatted_xml = g_strdup_printf("%s", xml);
	}
	
	printf("fixed whitespace xml:%s\n", formatted_xml);
	
	offset = 0;
	xmlTextReaderPtr reader;
	int ret;
	gtk_text_buffer_begin_user_action(textbuffer);
	reader = xmlReaderForMemory((char *)formatted_xml, strlen((char *)formatted_xml), NULL, "", 0);
	if (reader != NULL) {
		ret = xmlTextReaderRead(reader);
		while (ret == 1) {
			processHTMLNode(textbuffer, reader);
			ret = xmlTextReaderRead(reader);
		}
		xmlFreeTextReader(reader);
		if (ret != 0) {
			printf("Failed to parse xml\n");
		}
	} else {
		printf("Unable to open xml\n");
	}
	gtk_text_buffer_end_user_action(textbuffer);
	printf("\nNow finished reading...\n");
	// g_string_free(formatted_xml, FALSE);
	//g_free(xml);
}

void processHTMLNode(GtkTextBuffer * textbuffer, xmlTextReaderPtr reader)
{
	const xmlChar *name, *value;

	name = (const xmlChar*)g_utf8_strdown((const gchar*)xmlTextReaderConstName(reader), g_utf8_strlen((const gchar*)xmlTextReaderConstName(reader), -1));
	if (name == NULL)
		name = BAD_CAST "--";

	value = xmlTextReaderConstValue(reader);

	printf("%d %d %s %d %d",
	       xmlTextReaderDepth(reader),
	       xmlTextReaderNodeType(reader),
	       name,
	       xmlTextReaderIsEmptyElement(reader),
	       xmlTextReaderHasValue(reader));
	if (value == NULL)
		printf("\n");
	else {
		if (xmlStrlen(value) > 40)
			printf(" %.40s...\n", value);
		else
			printf(" %s\n", value);
	}

	int nodetype =  xmlTextReaderNodeType(reader);
	switch (nodetype)
	{
		case ((int)XML_READER_TYPE_ELEMENT):
			printf("type_element:%s\n", name);
			if (!xmlStrcasecmp(name, (const xmlChar*)"html"))
			{
				break;
			}
			if (!xmlStrcasecmp(name, (const xmlChar*)"img"))
			{
				printf("img tag found, offset:%d\n", GINT_TO_POINTER(offset));
				const gchar * attribute = (const gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"src");
				if(attribute != NULL)
				{
					//get pixbuf
					gchar image_encoded[strlen(attribute)];
					gchar*base64find = g_strstr_len (attribute, strlen(attribute), "base64,");
					strrepl(base64find, "base64,", "", image_encoded);
					//printf("image_encoded:%s\n", image_encoded);
					GdkPixbuf * mypixbuf = byoj_image_repr_read_b64 (image_encoded);
					//set file_string attached to pixbuf
					gdk_pixbuf_set_option(mypixbuf, "file_string", image_encoded);
					//Set iter
					GtkTextIter image_iter;
					gtk_text_buffer_get_iter_at_offset(textbuffer, &image_iter, offset);
					gtk_text_buffer_insert_pixbuf(textbuffer, &image_iter, mypixbuf);
					offset += 1;
				}
				break;
			}
			import_tag_stack = g_list_prepend (import_tag_stack, GINT_TO_POINTER(offset)); //push iter on to the list the beginning of the list
			printf("added iter for tag:%s\n", name);
			break;
		case ((int)XML_READER_TYPE_WHITESPACE):
			printf("type_whitespace:%s\n", name);
			break;
		case ((int)XML_READER_TYPE_TEXT):
			printf("type_text:%s\n", name);
			GtkTextIter insert_at;
			gtk_text_buffer_get_iter_at_offset(textbuffer, &insert_at, offset);
			gtk_text_buffer_insert(textbuffer, &insert_at, (const gchar*) value, xmlStrlen(value));
			offset += xmlStrlen(value);
			break;
		case((int)XML_READER_TYPE_ATTRIBUTE):
			break;
		case ((int)XML_READER_TYPE_END_ELEMENT):
			printf("type_end_element:%s\n", name);
			if (!xmlStrcasecmp(name, (const xmlChar*)"html"))
			{
				break;
			}
			int * offset_start = g_list_nth_data(import_tag_stack, 0); //pop off offset
			GtkTextIter begin_iter, end_iter;
			gtk_text_buffer_get_iter_at_offset(textbuffer, &begin_iter, GPOINTER_TO_INT(offset_start));
			gtk_text_buffer_get_iter_at_offset(textbuffer, &end_iter, offset);
			import_tag_stack = g_list_remove(import_tag_stack, offset_start); //remove offset

			GtkTextTag * tag;
			tag = gtk_text_tag_table_lookup(tagtable, (const gchar*) name);
			if(tag != NULL)
			{
				gtk_text_buffer_apply_tag(textbuffer, tag, &begin_iter, &end_iter);
			}
			if(xmlTextReaderHasAttributes(reader))
			{
				//font tag
				if(xmlStrstr(name, (xmlChar *)"font")!= NULL)
				{
					process_font_tag(textbuffer, reader, &begin_iter, &end_iter);
				}
				//CSS Style tags
				parse_and_process_style_attribute(textbuffer, reader, &begin_iter, &end_iter);
			}
			break;
		default:
			printf("other");
	}
}

int strrepl(char * InputString,char * StringToFind, char * StringToReplace,char * output)
{
	char *offset = NULL, *CurrentPointer = NULL;
	int insertlen;
	int findlen = strlen(StringToFind);
	int result = 0;

	if (StringToReplace)
		insertlen = strlen(StringToReplace);
	else
		insertlen = 0;
	if (output) {
		if (output != InputString)
			memmove(output,InputString,strlen(InputString)+1);
		InputString = output;
	}
	else
		result = strlen(InputString)+1;

	while (*InputString)        {
		offset = strstr(!offset?InputString:CurrentPointer,StringToFind);
		if (offset == NULL)
			break;
		CurrentPointer = (offset + (output ? insertlen : findlen));
		if (output) {
			strcpy (offset, (offset + findlen));
			memmove (offset + insertlen,
				 offset, strlen (offset) + 1);
			if (insertlen)
				memcpy (offset, StringToReplace, insertlen);
			result++;
		}
		else {
			result -= findlen;
			result += insertlen;
		}
	}
	return result;
}

void open_symbol_dialog (GtkTextBuffer * textbuffer)
{
	GtkWidget * dialogSymbol;
	dialogSymbol = create_dialogSymbol();
	
	GtkWidget *combobox_unicodetype = lookup_widget(GTK_WIDGET(dialogSymbol), "combobox_unicodetype");
	gtk_combo_box_set_active(GTK_COMBO_BOX(combobox_unicodetype), 2);
	on_combobox_unicodetype_changed(combobox_unicodetype);
	//gtk_widget_show(tableSymbol);
	if (gtk_dialog_run (GTK_DIALOG (dialogSymbol)) == GTK_RESPONSE_OK)
	{
		GtkWidget *label_Output = lookup_widget(GTK_WIDGET(dialogSymbol), "label_Output");
		const gchar * output_text = gtk_label_get_text(GTK_LABEL(label_Output));
		if(strcmp(output_text, "(none)"))
		{
			g_printf("ADDED:%s\n", output_text);
			gtk_text_buffer_insert_at_cursor(textbuffer, output_text, -1) ;
		}
		gtk_text_buffer_set_modified (textbuffer, TRUE);
	}
	gtk_widget_hide(dialogSymbol);
	
	gtk_widget_destroy (dialogSymbol);
}

void on_button_Symbol_activate (GtkButton * button, gpointer user_data)
{
	GtkWidget *label_Selected_Symbol = lookup_widget(GTK_WIDGET(button), "label_Selected_Symbol");
	const gchar * label =	gtk_button_get_label(button);
	gchar buf[40];
	g_sprintf(buf, "<span size=\"36000\">%s</span>\0", label);
	g_printf("ACTIVATED:%s\n", buf);
	gtk_label_set_markup (GTK_LABEL(label_Selected_Symbol), buf);
}

void on_button_Symbol_clicked (GtkButton * button, gpointer user_data)
{
	GtkWidget * label_Output = lookup_widget(GTK_WIDGET(button), "label_Output");
	const gchar * label =	gtk_button_get_label(button);
	const gchar * output_text = gtk_label_get_text(GTK_LABEL(label_Output));
	gchar buf[48];
	if(!strcmp(output_text, "(none)"))
	{
		g_sprintf(buf, "%s", label);
	}
	else
	{
		g_sprintf(buf, "%s%s", output_text, label);
	}
	g_printf("OUTPUT:%s\n", buf);
	gtk_label_set_text (GTK_LABEL(label_Output), buf);
}
